#' @include de_serialize.R
#' @include clustering.R
NULL

#' CLAWS: CLustering with wAvelets and Wer distanceS
#'
#' Groups electricity power curves (or any series of similar nature) by applying PAM
#' algorithm in parallel to chunks of size \code{nb_series_per_chunk}. Input series
#' must be sampled on the same time grid, no missing values.
#'
#' @param getSeries Access to the (time-)series, which can be of one of the three
#'   following types:
#'   \itemize{
#'     \item matrix: each line contains all the values for one time-serie, ordered by time
#'     \item connection: any R connection object (e.g. a file) providing lines as described above
#'     \item function: a custom way to retrieve the curves; it has only one argument:
#'       the indices of the series to be retrieved. See examples
#'   }
#' @param K1 Number of super-consumers to be found after stage 1 (K1 << N)
#' @param K2 Number of clusters to be found after stage 2 (K2 << K1)
#' @param random TRUE (default) for random chunks repartition
#' @param wf Wavelet transform filter; see ?wavelets::wt.filter. Default: haar
#' @param WER "end" to apply stage 2 after stage 1 has fully iterated, or "mix" to apply stage 2
#'   at the end of each task
#' @param ntasks Number of tasks (parallel iterations to obtain K1 medoids); default: 1.
#'   Note: ntasks << N, so that N is "roughly divisible" by N (number of series)
#' @param ncores_tasks "MPI" number of parallel tasks (1 to disable: sequential tasks)
#' @param ncores_clust "OpenMP" number of parallel clusterings in one task
#' @param nb_series_per_chunk (~Maximum) number of series in each group, inside a task
#' @param min_series_per_chunk Minimum number of series in each group
#' @param sep Separator in CSV input file (relevant only if getSeries is a file name)
#' @param nbytes Number of bytes to serialize a floating-point number; 4 or 8
#' @param endian Endianness to use for (de)serialization. Use "little" or "big" for portability
#'
#' @return A matrix of the final medoids curves
#'
#' @examples
#' getData = function(start, n) {
#'   con = dbConnect(drv = RSQLite::SQLite(), dbname = "mydata.sqlite")
#'   df = dbGetQuery(con, paste(
#'     "SELECT * FROM times_values GROUP BY id OFFSET ",start,
#'     "LIMIT ", n, " ORDER BY date", sep=""))
#'   return (df)
#' }
#' #####TODO: if DB, array rank --> ID at first retrieval, when computing coeffs; so:: NO use of IDs !
#'   #TODO: 3 examples, data.frame / binary file / DB sqLite
#'   + sampleCurves : wavBootstrap de package wmtsa
#' cl = epclust(getData, K1=200, K2=15, ntasks=1000, nb_series_per_chunk=5000, WER="mix")
#' @export
claws = function(getSeries, K1, K2,
	random=TRUE, #randomize series order?
	wf="haar", #stage 1
	WER="end", #stage 2
	ntasks=1, ncores_tasks=1, ncores_clust=4, #control parallelism
	nb_series_per_chunk=50*K1, min_series_per_chunk=5*K1, #chunk size
	sep=",", #ASCII input separator
	nbytes=4, endian=.Platform$endian) #serialization (write,read)
{
	# Check/transform arguments
	if (!is.matrix(getSeries) && !is.function(getSeries) &&
		!is(getSeries, "connection" && !is.character(getSeries)))
	{
		stop("'getSeries': matrix, function, file or valid connection (no NA)")
	}
	K1 = .toInteger(K1, function(x) x>=2)
	K2 = .toInteger(K2, function(x) x>=2)
	if (!is.logical(random))
		stop("'random': logical")
	tryCatch(
		{ignored <- wt.filter(wf)},
		error = function(e) stop("Invalid wavelet filter; see ?wavelets::wt.filter"))
	if (WER!="end" && WER!="mix")
		stop("WER takes values in {'end','mix'}")
	ntasks = .toInteger(ntasks, function(x) x>=1)
	ncores_tasks = .toInteger(ncores_tasks, function(x) x>=1)
	ncores_clust = .toInteger(ncores_clust, function(x) x>=1)
	nb_series_per_chunk = .toInteger(nb_series_per_chunk, function(x) x>=K1)
	min_series_per_chunk = .toInteger(K1, function(x) x>=K1 && x<=nb_series_per_chunk)
	if (!is.character(sep))
		stop("'sep': character")
	nbytes = .toInteger(nbytes, function(x) x==4 || x==8)

	# Serialize series if required, to always use a function
	bin_dir = "epclust.bin/"
	dir.create(bin_dir, showWarnings=FALSE, mode="0755")
	if (!is.function(getSeries))
	{
		series_file = paste(bin_dir,"data",sep="") ; unlink(series_file)
		serialize(getSeries, series_file, nb_series_per_chunk, sep, nbytes, endian)
		getSeries = function(indices) getDataInFile(indices, series_file, nbytes, endian)
	}

	# Serialize all wavelets coefficients (+ IDs) onto a file
	coefs_file = paste(bin_dir,"coefs",sep="") ; unlink(coefs_file)
	index = 1
	nb_curves = 0
	repeat
	{
		series = getSeries((index-1)+seq_len(nb_series_per_chunk))
		if (is.null(series))
			break
		coefs_chunk = curvesToCoefs(series, wf)
		serialize(coefs_chunk, coefs_file, nb_series_per_chunk, sep, nbytes, endian)
		index = index + nb_series_per_chunk
		nb_curves = nb_curves + nrow(coefs_chunk)
	}
	getCoefs = function(indices) getDataInFile(indices, coefs_file, nbytes, endian)

	if (nb_curves < min_series_per_chunk)
		stop("Not enough data: less rows than min_series_per_chunk!")
	nb_series_per_task = round(nb_curves / ntasks)
	if (nb_series_per_task < min_series_per_chunk)
		stop("Too many tasks: less series in one task than min_series_per_chunk!")

	# Cluster coefficients in parallel (by nb_series_per_chunk)
	indices_all = if (random) sample(nb_curves) else seq_len(nb_curves)
	indices_tasks = lapply(seq_len(ntasks), function(i) {
		upper_bound = ifelse( i<ntasks, min(nb_series_per_task*i,nb_curves), nb_curves )
		indices_all[((i-1)*nb_series_per_task+1):upper_bound]
	})
	cl = parallel::makeCluster(ncores_tasks)
	# 1000*K1 indices [if WER=="end"], or empty vector [if WER=="mix"] --> series on file
	indices = unlist( parallel::parLapply(cl, indices_tasks, function(inds) {
		indices_medoids = clusteringTask(inds,getCoefs,K1,nb_series_per_chunk,ncores_clust)
		if (WER=="mix")
		{
			medoids2 = computeClusters2(
				getSeries(indices_medoids), K2, getSeries, nb_series_per_chunk)
			serialize(medoids2, synchrones_file, nb_series_per_chunk, sep, nbytes, endian)
			return (vector("integer",0))
		}
		indices_medoids
	}) )
	parallel::stopCluster(cl)

	getRefSeries = getSeries
	synchrones_file = paste(bin_dir,"synchrones",sep="") ; unlink(synchrones_file)
	if (WER=="mix")
	{
		indices = seq_len(ntasks*K2)
		#Now series must be retrieved from synchrones_file
		getSeries = function(inds) getDataInFile(inds, synchrones_file, nbytes, endian)
		#Coefs must be re-computed
		unlink(coefs_file)
		index = 1
		repeat
		{
			series = getSeries((index-1)+seq_len(nb_series_per_chunk))
			if (is.null(series))
				break
			coefs_chunk = curvesToCoefs(series, wf)
			serialize(coefs_chunk, coefs_file, nb_series_per_chunk, sep, nbytes, endian)
			index = index + nb_series_per_chunk
		}
	}

	# Run step2 on resulting indices or series (from file)
	indices_medoids = clusteringTask(
		indices, getCoefs, K1, nb_series_per_chunk, ncores_tasks*ncores_clust)
	computeClusters2(getSeries(indices_medoids),K2,getRefSeries,nb_series_per_chunk)
}

# helper
curvesToCoefs = function(series, wf)
{
	L = length(series[1,])
	D = ceiling( log2(L) )
	nb_sample_points = 2^D
	t( apply(series, 1, function(x) {
		interpolated_curve = spline(1:L, x, n=nb_sample_points)$y
		W = wavelets::dwt(interpolated_curve, filter=wf, D)@W
		rev( sapply( W, function(v) ( sqrt( sum(v^2) ) ) ) )
	}) )
}

# helper
.toInteger <- function(x, condition)
{
	if (!is.integer(x))
		tryCatch(
			{x = as.integer(x)[1]},
			error = function(e) paste("Cannot convert argument",substitute(x),"to integer")
		)
	if (!condition(x))
		stop(paste("Argument",substitute(x),"does not verify condition",body(condition)))
	x
}
